@@ -9,7 +9,7 @@
 #  (      _ \     /  |     (   | (_ |    |      |
 # \___| _/  _\ _|_\ ____| \___/ \___|   _|     _|
 
-# __init__.py -> carlogtt_python_library
+# src/carlogtt_library/__init__.py
 # Created 10/4/23 - 10:44 AM UK Time (London) by carlogtt
 # Copyright (c) Amazon.com Inc. All Rights Reserved.
 # AMAZON.COM CONFIDENTIAL
@@ -53,6 +53,7 @@ from .utils import *
 
 # Setting up logger for current module
 _module_logger = _logging.getLogger(__name__)
+_module_logger.addHandler(_logging.NullHandler())
 
 
 class _CompatibilityProxy:
@@ -150,8 +151,8 @@ class _CompatibilityProxy:
     def _handle_cli_style(self, name: str):
         upper_name = name.upper()
         msg = (
-            f"[DEPRECATED] '{name}' is deprecated. Use '{CLIStyle.__qualname__}.{upper_name}'"
-            " instead."
+            f"[DEPRECATED] '{name}' is deprecated in package '{__package__}'. Use"
+            f" '{CLIStyle.__qualname__}.{upper_name}' instead."
         )
 
         _warnings.warn(msg, DeprecationWarning, stacklevel=3)
@@ -161,7 +162,7 @@ class _CompatibilityProxy:
 
     def _warn_user(self, name: str, new_class: type):
         msg = (
-            f"[DEPRECATED] '{name}' is deprecated. Use the parent class"
+            f"[DEPRECATED] '{name}' is deprecated in package '{__package__}'. Use the parent class"
             f" '{new_class.__qualname__}()' instead."
         )
 
