@@ -9,7 +9,7 @@
 #  (      _ \     /  |     (   | (_ |    |      |
 # \___| _/  _\ _|_\ ____| \___/ \___|   _|     _|
 
-# simt.py
+# src/carlogtt_library/amazon_internal/simt.py
 # Created 10/25/23 - 3:55 PM UK Time (London) by carlogtt
 # Copyright (c) Amazon.com Inc. All Rights Reserved.
 # AMAZON.COM CONFIDENTIAL
@@ -33,6 +33,8 @@ This module ...
 
 # Standard Library Imports
 import functools
+import logging
+import warnings
 from collections.abc import Generator
 from typing import Any, Optional
 
@@ -50,11 +52,18 @@ from .. import exceptions, utils
 
 # List of public names in the module
 __all__ = [
+    'SimT',
     'SimTicketHandler',
 ]
 
+# Setting up logger for current module
+module_logger = logging.getLogger(__name__)
 
-class SimTicketHandler:
+# Type aliases
+TicketyClient = Any
+
+
+class SimT:
     """
     A handler class for the TicketyPythonSdk.
 
@@ -104,17 +113,15 @@ class SimTicketHandler:
         self,
         aws_region_name: str,
         *,
-        aws_account_id: str = "default",
-        ticketing_system_name: str = "default",
         aws_profile_name: Optional[str] = None,
         aws_access_key_id: Optional[str] = None,
         aws_secret_access_key: Optional[str] = None,
         aws_session_token: Optional[str] = None,
         caching: bool = False,
         client_parameters: Optional[dict[str, Any]] = None,
+        aws_account_id: str = "default",
+        ticketing_system_name: str = "default",
     ) -> None:
-        self._aws_account_id = aws_account_id
-        self._ticketing_system_name = ticketing_system_name
         self._aws_region_name = aws_region_name
         self._aws_profile_name = aws_profile_name
         self._aws_access_key_id = aws_access_key_id
@@ -122,12 +129,15 @@ class SimTicketHandler:
         self._aws_session_token = aws_session_token
         self._caching = caching
         self._cache: dict[str, Any] = dict()
+        self._aws_account_id = aws_account_id
+        self._ticketing_system_name = ticketing_system_name
+        self._aws_tickety_region_name = "global"
         self._aws_service_name = "tickety"
         self._aws_endpoint_url = "https://global.api.tickety.amazon.dev"
         self._client_parameters = client_parameters if client_parameters else dict()
 
     @property
-    def _client(self):
+    def _client(self) -> TicketyClient:
         if self._caching:
             if self._cache.get('client') is None:
                 self._cache['client'] = self._get_boto_tickety_client()
@@ -163,17 +173,17 @@ class SimTicketHandler:
 
         return botocore.config.Config(signature_version="v4a", retries={"mode": "adaptive"})
 
-    def _get_boto_tickety_client(self):
+    def _get_boto_tickety_client(self) -> TicketyClient:
         """
         Create a low level tickety client.
 
         :return: A tickety client.
-        :raise: SimTHandlerError if function call fails.
+        :raise: SimTError if function call fails.
         """
 
         try:
             boto_session = boto3.session.Session(
-                region_name="global",  # overriding region with 'global'
+                region_name=self._aws_tickety_region_name,
                 profile_name=self._aws_profile_name,
                 aws_access_key_id=self._aws_access_key_id,
                 aws_secret_access_key=self._aws_secret_access_key,
@@ -189,10 +199,32 @@ class SimTicketHandler:
             return client
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response))
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex))
+
+    def invalidate_client_cache(self) -> None:
+        """
+        Clears the cached client, if caching is enabled.
+
+        This method allows manually invalidating the cached client,
+        forcing a new client instance to be created on the next access.
+        Useful if AWS credentials have changed or if there's a need to
+        connect to a different region within the same instance
+        lifecycle.
+
+        :return: None.
+        :raise SimTError: Raises an error if caching is not enabled
+               for this instance.
+        """
+
+        if not self._cache:
+            raise exceptions.SimTError(
+                f"Session caching is not enabled for this instance of {self.__class__.__qualname__}"
+            )
+
+        self._cache['client'] = None
 
     def get_tickets(self, filters: dict[str, Any]) -> Generator[dict[str, Any], None, None]:
         """
@@ -200,23 +232,21 @@ class SimTicketHandler:
 
         :param filters:
         :return: Returns an iterable of ticketIds.
-        :raise: SimTHandlerError if function call fails.
+        :raise: SimTError if function call fails.
         """
 
-        @utils.retry(Exception)
-        def _list_tickets_with_retry(payload: dict[str, Any]):
-            return self._client.list_tickets(
-                awsAccountId=self._aws_account_id,
-                ticketingSystemName=self._ticketing_system_name,
-                **payload,
-            )
-
         payload: dict[str, Any] = {'filters': filters}
         next_token = 'valid_string'
 
         try:
             while next_token:
-                tickety_response = _list_tickets_with_retry(payload)
+                with utils.retry(exception_to_check=Exception) as retryer:
+                    tickety_response = retryer(
+                        self._client.list_tickets,
+                        awsAccountId=self._aws_account_id,
+                        ticketingSystemName=self._ticketing_system_name,
+                        **payload,
+                    )
 
                 next_token = tickety_response.get('nextToken', '')
                 payload.update({'nextToken': next_token})
@@ -225,12 +255,12 @@ class SimTicketHandler:
                     yield ticket_data
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex)) from None
 
-    @utils.retry(exceptions.SimTHandlerError)
+    @utils.retry(exception_to_check=exceptions.SimTError)
     def get_ticket_details(
         self,
         ticket_id: str,
@@ -243,7 +273,7 @@ class SimTicketHandler:
 
         :param ticket_id: The ticket ID.
         :return:
-        :raise: SimTHandlerError if function call fails.
+        :raise: SimTError if function call fails.
         """
 
         try:
@@ -257,15 +287,15 @@ class SimTicketHandler:
                 return tickety_response['ticket']
 
             except KeyError as ex_inner:
-                raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex_inner)}")
+                raise exceptions.SimTError(str(ex_inner))
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex)) from None
 
-    @utils.retry(exceptions.SimTHandlerError)
+    @utils.retry(exception_to_check=exceptions.SimTError)
     def update_ticket(self, ticket_id: str, payload: dict[str, Any]) -> None:
         """
         Updates a ticket in the ticketing system.
@@ -276,7 +306,7 @@ class SimTicketHandler:
         :param ticket_id: The ticket ID.
         :param payload:
         :return:
-        :raise: SimTHandlerError if function call fails.
+        :raise: SimTError if function call fails.
         """
 
         try:
@@ -288,18 +318,18 @@ class SimTicketHandler:
             )
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex)) from None
 
         if (
             not isinstance(response, dict)
             or response.get('ResponseMetadata', {}).get('HTTPStatusCode') != 200
         ):
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(response)}")
+            raise exceptions.SimTError(str(response))
 
-    @utils.retry(exceptions.SimTHandlerError)
+    @utils.retry(exception_to_check=exceptions.SimTError)
     def create_ticket_comment(
         self,
         ticket_id: str,
@@ -323,7 +353,7 @@ class SimTicketHandler:
                "text/amz-markdown-sim" "text/plain".
                Auto default to "text/amz-markdown-sim".
         :return: The SIM-T commentId as a string.
-        :raise: SimTHandlerError if function call fails.
+        :raise: SimTError if function call fails.
         """
 
         payload = {
@@ -344,15 +374,15 @@ class SimTicketHandler:
                 return tickety_response['commentId']
 
             except KeyError as ex_inner:
-                raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex_inner)}")
+                raise exceptions.SimTError(str(ex_inner))
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex)) from None
 
-    @utils.retry(exceptions.SimTHandlerError)
+    @utils.retry(exception_to_check=exceptions.SimTError)
     def create_ticket(self, ticket_data: dict[str, Any]) -> str:
         """
         Create a ticket in the ticketing system and return the
@@ -373,10 +403,28 @@ class SimTicketHandler:
                 return ticket_response['id']
 
             except KeyError as ex_inner:
-                raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex_inner)}")
+                raise exceptions.SimTError(str(ex_inner))
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SimTError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SimTHandlerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SimTError(str(ex)) from None
+
+
+class SimTicketHandler(SimT):
+    """
+    DEPRECATED: Please use SimT instead.
+    This subclass only exists for backward compatibility.
+    """
+
+    def __init__(self, *args, **kwargs):
+        msg = (
+            f"[DEPRECATED] '{__package__}' class 'SimTicketHandler' is deprecated. Use 'SimT'"
+            " instead."
+        )
+
+        warnings.warn(msg, DeprecationWarning, stacklevel=2)
+        module_logger.warning(msg)
+
+        super().__init__(*args, **kwargs)
