@@ -9,7 +9,7 @@
 #  (      _ \     /  |     (   | (_ |    |      |
 # \___| _/  _\ _|_\ ____| \___/ \___|   _|     _|
 
-# database_sql.py
+# src/carlogtt_library/database/database_sql.py
 # Created 9/25/23 - 2:34 PM UK Time (London) by carlogtt
 # Copyright (c) Amazon.com Inc. All Rights Reserved.
 # AMAZON.COM CONFIDENTIAL
@@ -33,10 +33,13 @@ This module ...
 
 # Standard Library Imports
 import abc
+import datetime
+import decimal
 import logging
 import pathlib
 import sqlite3
-from collections.abc import Generator
+import time
+from collections.abc import Generator, Iterable, Sequence
 from typing import Any, Optional, Union
 
 # Third Party Library Imports
@@ -47,6 +50,7 @@ from mysql.connector.pooling import PooledMySQLConnection
 
 # Local Folder (Relative) Imports
 from .. import exceptions, utils
+from . import database_utils
 
 # psycopg2 is defined as an optional dependency. We use this try/except
 # to gracefully handle environments where psycopg2 is not installed
@@ -89,10 +93,26 @@ module_logger = logging.getLogger(__name__)
 
 # Type aliases
 MySQLConn = Union[MySQLConnectionAbstract, PooledMySQLConnection]
-SQLValueType = Any
+SQLValueType = Union[
+    bool,
+    bytes,
+    float,
+    int,
+    str,
+    decimal.Decimal,
+    datetime.date,
+    datetime.time,
+    datetime.datetime,
+    datetime.timedelta,
+    time.struct_time,
+    None,
+]
 
 
 class Database(abc.ABC):
+
+    db_utils: database_utils.DatabaseUtils
+
     @abc.abstractmethod
     def open_db_connection(self) -> None:
         pass
@@ -102,14 +122,18 @@ class Database(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def send_to_db(self, sql_query: str, sql_values: Union[tuple[SQLValueType, ...], str]) -> None:
+    def send_to_db(self, sql_query: str, sql_values: Sequence[SQLValueType] = ()) -> None:
+        pass
+
+    @abc.abstractmethod
+    def send_many_to_db(self, sql_query: str, sql_values: Iterable[Sequence[SQLValueType]]) -> None:
         pass
 
     @abc.abstractmethod
     def fetch_from_db(
         self,
         sql_query: str,
-        sql_values: Union[tuple[SQLValueType, ...], str],
+        sql_values: Sequence[SQLValueType] = (),
         *,
         fetch_one: bool = False,
     ) -> Generator[dict[str, Any], None, None]:
@@ -125,6 +149,12 @@ class MySQL(Database):
     :param password: Password to authenticate with the MySQL server.
     :param port: Port number of the MySQL server.
     :param database_schema: Name of the database schema to use.
+
+    **Attributes**
+
+    ``db_utils``
+        Instance of :class:`~database_utils.DatabaseUtils`
+        (helper for reading external SQL files, etc.).
     """
 
     def __init__(self, host: str, user: str, password: str, port: str, database_schema: str):
@@ -134,6 +164,7 @@ class MySQL(Database):
         self._port = port
         self._database_schema = database_schema
         self._db_connection: Optional[MySQLConn] = None
+        self.db_utils = database_utils.DatabaseUtils()
 
     @property
     def db_connection(self) -> MySQLConn:
@@ -151,7 +182,7 @@ class MySQL(Database):
 
         return self._db_connection
 
-    @utils.retry(exceptions.MySQLError)
+    @utils.retry(exception_to_check=exceptions.MySQLError)
     def open_db_connection(self) -> None:
         """
         Open a MySQL db connection.
@@ -170,11 +201,13 @@ class MySQL(Database):
             )
 
         except mysql.connector.Error as ex:
-            message = f"While connecting to [{self._host}] operation failed! traceback: {repr(ex)}"
+            message = (
+                f"While connecting to host [{self._host}] operation failed! traceback: {repr(ex)}"
+            )
             module_logger.error(message)
-            raise exceptions.MySQLError(message)
+            raise exceptions.MySQLError(message) from None
 
-    @utils.retry(exceptions.MySQLError)
+    @utils.retry(exception_to_check=exceptions.MySQLError)
     def close_db_connection(self) -> None:
         """
         Close the MySQL db connection.
@@ -184,18 +217,16 @@ class MySQL(Database):
         """
 
         try:
-            assert isinstance(self._db_connection, MySQLConnectionAbstract) or isinstance(
-                self._db_connection, PooledMySQLConnection
-            ), "Database connection is not open!"
-
-            self._db_connection.close()
+            if self._db_connection:
+                self._db_connection.close()
+                self._db_connection = None
 
         except mysql.connector.Error as ex:
             message = f"While closing [{self._host}] operation failed! traceback: {repr(ex)}"
             module_logger.error(message)
-            raise exceptions.MySQLError(message)
+            raise exceptions.MySQLError(message) from None
 
-    def send_to_db(self, sql_query: str, sql_values: Union[tuple[SQLValueType, ...], str]) -> None:
+    def send_to_db(self, sql_query: str, sql_values: Sequence[SQLValueType] = ()) -> None:
         """
         Send data to MySQL database.
 
@@ -207,30 +238,68 @@ class MySQL(Database):
         db_cursor = self.db_connection.cursor(prepared=True)
 
         try:
-            db_cursor.execute(sql_query, sql_values)
+            with utils.retry(exception_to_check=Exception) as retryer:
+                retryer(db_cursor.execute, sql_query, sql_values)
+                retryer(self.db_connection.commit)
 
-            self.db_connection.commit()
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
-
-        except mysql.connector.OperationalError as ex:
+        except mysql.connector.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._host}] operation failed!"
+                f"While executing SQL query {sql_query=} on host [{self._host}] operation failed!"
                 f" traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.MySQLError(message)
+            raise exceptions.MySQLError(message) from None
 
         finally:
             db_cursor.close()
             self.close_db_connection()
 
+    def send_many_to_db(self, sql_query: str, sql_values: Iterable[Sequence[SQLValueType]]) -> None:
+        """
+        Execute the same SQL statement many times in a single
+        ACID‑compliant transaction. Commit only if every execution
+        succeeds, otherwise roll back.
+
+        :param sql_query: The parametrized SQL string.
+        :param sql_values: Any iterable yielding values to be
+            substituted in the SQL query.
+        :raise MySQLError: (after rollback) If the operation fails.
+        """
+
+        db_cursor = self.db_connection.cursor(prepared=True)
+
+        try:
+            # executemany sends the whole batch; server handles each row
+            # list() ensures we don’t exhaust a generator if retries
+            with utils.retry(exception_to_check=Exception) as retryer:
+                retryer(db_cursor.executemany, sql_query, list(sql_values))
+                retryer(self.db_connection.commit)
+
+            module_logger.debug(
+                "Database atomic batch executed and committed successfully: "
+                f"{db_cursor.rowcount} rows affected by SQL query {sql_query=}"
+            )
+
+        except mysql.connector.Error as ex:
+            # Atomicity guarantee
+            self.db_connection.rollback()
+
+            message = (
+                f"Database atomic batch for SQL query {sql_query} failed on host [{self._host}]. "
+                f"Rolled back the entire transaction. Traceback: {repr(ex)}"
+            )
+            module_logger.error(message)
+            raise exceptions.MySQLError(message) from None
+
+        finally:
+            db_cursor.close()
+            self.close_db_connection()
+
+    @utils.retry(exception_to_check=exceptions.MySQLError, delay_secs=2)
     def fetch_from_db(
-        self,
-        sql_query: str,
-        sql_values: Union[tuple[SQLValueType, ...], str],
-        *,
-        fetch_one: bool = False,
+        self, sql_query: str, sql_values: Sequence[SQLValueType] = (), *, fetch_one: bool = False
     ) -> Generator[dict[str, Any], None, None]:
         """
         Fetch data from MySQL database.
@@ -248,41 +317,37 @@ class MySQL(Database):
         try:
             db_cursor.execute(sql_query, sql_values)
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
             if fetch_one:
                 next_row = db_cursor.fetchone()
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield next_row
 
             else:
                 next_row = db_cursor.fetchone()
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield next_row
                     # db_cursor.fetchone will return None when at the
                     # end so the sentinel is met by the iter function
                     yield from iter(db_cursor.fetchone, None)
 
-        except (mysql.connector.OperationalError, mysql.connector.errors.ProgrammingError) as ex:
+        except mysql.connector.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._host}] operation failed!"
+                f"While executing SQL query {sql_query=} on host [{self._host}] operation failed!"
                 f" traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.MySQLError(message)
+            raise exceptions.MySQLError(message) from None
 
         finally:
-            # cursor.reset typically discards the results of the last
-            # query and resets the cursor to its initial state, without
-            # affecting the underlying database connection. This is
-            # useful if you've fetched some rows from a result but want
-            # to discard the remaining unfetched rows and reuse the
-            # cursor for another query.
-            db_cursor.reset()
+            db_cursor.close()
             self.close_db_connection()
 
 
@@ -295,6 +360,12 @@ class PostgreSQL(Database):
     :param password: Password to authenticate with the Postgres server.
     :param port: Port number of the Postgres server.
     :param database_schema: Name of the database schema to use.
+
+    **Attributes**
+
+    ``db_utils``
+        Instance of :class:`~database_utils.DatabaseUtils`
+        (helper for reading external SQL files, etc.).
     """
 
     def __init__(self, host: str, user: str, password: str, port: str, database_schema: str):
@@ -304,6 +375,7 @@ class PostgreSQL(Database):
         self._port = port
         self._database_schema = database_schema
         self._db_connection: Optional[PGConnection] = None
+        self.db_utils = database_utils.DatabaseUtils()
 
     @property
     def db_connection(self) -> PGConnection:
@@ -321,7 +393,7 @@ class PostgreSQL(Database):
 
         return self._db_connection
 
-    @utils.retry(exceptions.PostgresError)
+    @utils.retry(exception_to_check=exceptions.PostgresError)
     def open_db_connection(self) -> None:
         """
         Open a PostgreSQL db connection.
@@ -339,12 +411,14 @@ class PostgreSQL(Database):
                 port=self._port,
             )
 
-        except psycopg2.OperationalError as ex:
-            message = f"While connecting to [{self._host}] operation failed! traceback: {repr(ex)}"
+        except psycopg2.Error as ex:
+            message = (
+                f"While connecting to host [{self._host}] operation failed! traceback: {repr(ex)}"
+            )
             module_logger.error(message)
-            raise exceptions.PostgresError(message)
+            raise exceptions.PostgresError(message) from None
 
-    @utils.retry(exceptions.PostgresError)
+    @utils.retry(exception_to_check=exceptions.PostgresError)
     def close_db_connection(self) -> None:
         """
         Close the PostgreSQL db connection.
@@ -361,10 +435,9 @@ class PostgreSQL(Database):
         except psycopg2.Error as ex:
             message = f"While closing [{self._host}] operation failed! traceback: {repr(ex)}"
             module_logger.error(message)
-            raise exceptions.PostgresError(message)
+            raise exceptions.PostgresError(message) from None
 
-    @utils.retry(exceptions.PostgresError, tries=3, delay_secs=2)
-    def send_to_db(self, sql_query: str, sql_values: Union[tuple, str]) -> None:
+    def send_to_db(self, sql_query: str, sql_values: Sequence[SQLValueType] = ()) -> None:
         """
         Send data to PostgreSQL database.
 
@@ -376,30 +449,68 @@ class PostgreSQL(Database):
         db_cursor = self.db_connection.cursor()
 
         try:
-            db_cursor.execute(sql_query, sql_values)
-            self.db_connection.commit()
+            with utils.retry(exception_to_check=Exception) as retryer:
+                retryer(db_cursor.execute, sql_query, sql_values)
+                retryer(self.db_connection.commit)
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
         except psycopg2.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._host}] "
+                f"While executing SQL query {sql_query=} on host [{self._host}] "
                 f"operation failed! traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.PostgresError(message)
+            raise exceptions.PostgresError(message) from None
 
         finally:
             db_cursor.close()
             self.close_db_connection()
 
-    @utils.retry(exceptions.PostgresError, tries=3, delay_secs=2)
+    def send_many_to_db(self, sql_query: str, sql_values: Iterable[Sequence[SQLValueType]]) -> None:
+        """
+        Execute the same SQL statement many times in a single
+        ACID‑compliant transaction. Commit only if every execution
+        succeeds, otherwise roll back.
+
+        :param sql_query: The parametrized SQL string.
+        :param sql_values: Any iterable yielding values to be
+            substituted in the SQL query.
+        :raise PostgresError: (after rollback) If the operation fails.
+        """
+
+        db_cursor = self.db_connection.cursor()
+
+        try:
+            # executemany sends the whole batch; server handles each row
+            # list() ensures we don’t exhaust a generator if retries
+            with utils.retry(exception_to_check=Exception) as retryer:
+                retryer(db_cursor.executemany, sql_query, list(sql_values))
+                retryer(self.db_connection.commit)
+
+            module_logger.debug(
+                "Database atomic batch executed and committed successfully: "
+                f"{db_cursor.rowcount} rows affected by SQL query {sql_query=}"
+            )
+
+        except psycopg2.Error as ex:
+            # Atomicity guarantee
+            self.db_connection.rollback()
+
+            message = (
+                f"Database atomic batch for SQL query {sql_query} failed on host [{self._host}]. "
+                f"Rolled back the entire transaction. Traceback: {repr(ex)}"
+            )
+            module_logger.error(message)
+            raise exceptions.PostgresError(message) from None
+
+        finally:
+            db_cursor.close()
+            self.close_db_connection()
+
+    @utils.retry(exception_to_check=exceptions.PostgresError, delay_secs=2)
     def fetch_from_db(
-        self,
-        sql_query: str,
-        sql_values: Union[tuple, str],
-        *,
-        fetch_one: bool = False,
+        self, sql_query: str, sql_values: Sequence[SQLValueType] = (), *, fetch_one: bool = False
     ) -> Generator[dict[str, Any], None, None]:
         """
         Fetch data from PostgreSQL database.
@@ -417,19 +528,21 @@ class PostgreSQL(Database):
         try:
             db_cursor.execute(sql_query, sql_values)
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
             if fetch_one:
                 next_row = db_cursor.fetchone()
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield dict(next_row)
 
             else:
                 next_row = db_cursor.fetchone()
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield dict(next_row)
                     # Fetch the rest one by one until None is returned
@@ -437,11 +550,11 @@ class PostgreSQL(Database):
 
         except psycopg2.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._host}] operation failed!"
+                f"While executing SQL query {sql_query=} on host [{self._host}] operation failed!"
                 f" traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.PostgresError(message)
+            raise exceptions.PostgresError(message) from None
 
         finally:
             db_cursor.close()
@@ -454,12 +567,19 @@ class SQLite(Database):
 
     :param sqlite_db_path: Fullpath to the SQLite database file.
     :param filename: Name of the SQLite database file.
+
+    **Attributes**
+
+    ``db_utils``
+        Instance of :class:`~database_utils.DatabaseUtils`
+        (helper for reading external SQL files, etc.).
     """
 
     def __init__(self, sqlite_db_path: Union[str, pathlib.Path], filename: str):
         self._sqlite_db_path = sqlite_db_path
         self._filename = filename
         self._db_connection: Optional[sqlite3.Connection] = None
+        self.db_utils = database_utils.DatabaseUtils()
 
     @property
     def db_connection(self) -> sqlite3.Connection:
@@ -499,12 +619,13 @@ class SQLite(Database):
             # at runtime using the PRAGMA command
             self._db_connection.execute("PRAGMA foreign_keys = ON;")
 
-        except sqlite3.OperationalError as ex:
+        except sqlite3.Error as ex:
             message = (
-                f"While connecting to [{self._filename}] operation failed! traceback: {repr(ex)}"
+                f"While connecting to host [{self._filename}] operation failed! traceback:"
+                f" {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.SQLiteError(message)
+            raise exceptions.SQLiteError(message) from None
 
     def close_db_connection(self) -> None:
         """
@@ -514,19 +635,16 @@ class SQLite(Database):
         """
 
         try:
-            assert isinstance(
-                self._db_connection, sqlite3.Connection
-            ), "Database connection is not open!"
+            if self._db_connection:
+                self._db_connection.close()
+                self._db_connection = None
 
-            self._db_connection.close()
-            self._db_connection = None
-
-        except sqlite3.OperationalError as ex:
+        except sqlite3.Error as ex:
             message = f"While closing [{self._filename}] operation failed! traceback: {repr(ex)}"
             module_logger.error(message)
-            raise exceptions.SQLiteError(message)
+            raise exceptions.SQLiteError(message) from None
 
-    def send_to_db(self, sql_query: str, sql_values: Union[tuple[SQLValueType, ...], str]) -> None:
+    def send_to_db(self, sql_query: str, sql_values: Sequence[SQLValueType] = ()) -> None:
         """
         Send data to SQLite database.
 
@@ -540,28 +658,65 @@ class SQLite(Database):
         try:
             db_cursor.execute(sql_query, sql_values)
 
-            db_cursor.connection.commit()
+            self.db_connection.commit()
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
-        except (sqlite3.OperationalError, sqlite3.IntegrityError) as ex:
+        except sqlite3.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._filename}] operation failed!"
-                f" traceback: {repr(ex)}"
+                f"While executing SQL query {sql_query=} on host [{self._filename}] operation"
+                f" failed! traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.SQLiteError(message)
+            raise exceptions.SQLiteError(message) from None
+
+        finally:
+            db_cursor.close()
+            self.close_db_connection()
+
+    def send_many_to_db(self, sql_query: str, sql_values: Iterable[Sequence[SQLValueType]]) -> None:
+        """
+        Execute the same SQL statement many times in a single
+        ACID‑compliant transaction. Commit only if every execution
+        succeeds, otherwise roll back.
+
+        :param sql_query: The parametrized SQL string.
+        :param sql_values: Any iterable yielding values to be
+            substituted in the SQL query.
+        :raise SQLiteError: (after rollback) If the operation fails.
+        """
+
+        db_cursor = self.db_connection.cursor()
+
+        try:
+            # executemany sends the whole batch; server handles each row
+            # list() ensures we don’t exhaust a generator if retries
+            db_cursor.executemany(sql_query, list(sql_values))
+
+            self.db_connection.commit()
+
+            module_logger.debug(
+                "Database atomic batch executed and committed successfully: "
+                f"{db_cursor.rowcount} rows affected by SQL query {sql_query=}"
+            )
+
+        except sqlite3.Error as ex:
+            # Atomicity guarantee
+            self.db_connection.rollback()
+
+            message = (
+                f"Database atomic batch for SQL query {sql_query} failed on host"
+                f" [{self._filename}]. Rolled back the entire transaction. Traceback: {repr(ex)}"
+            )
+            module_logger.error(message)
+            raise exceptions.SQLiteError(message) from None
 
         finally:
             db_cursor.close()
             self.close_db_connection()
 
     def fetch_from_db(
-        self,
-        sql_query: str,
-        sql_values: Union[tuple[SQLValueType, ...], str],
-        *,
-        fetch_one: bool = False,
+        self, sql_query: str, sql_values: Sequence[SQLValueType] = (), *, fetch_one: bool = False
     ) -> Generator[dict[str, Any], None, None]:
         """
         Fetch data from SQLite database.
@@ -573,52 +728,54 @@ class SQLite(Database):
         :raise SQLiteError: If the operation fails.
         """
 
-        def _next_row_dict() -> Optional[dict[str, Any]]:
-            """
-            Returns the next row as a dictionary or None if no
-            more rows.
-
-            :return: The next row as a dictionary or None.
-            """
-
-            row_fetched = db_cursor.fetchone()
-            if row_fetched is None:
-                return None
-            else:
-                return dict(row_fetched)
-
         db_cursor = self.db_connection.cursor()
 
         try:
             db_cursor.execute(sql_query, sql_values)
 
-            module_logger.info(f"Database SQL query {sql_query=} executed successfully")
+            module_logger.debug(f"Database SQL query {sql_query=} executed successfully")
 
             if fetch_one:
-                next_row = _next_row_dict()
+                next_row = self._next_row_dict(db_cursor=db_cursor)
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield next_row
 
             else:
-                next_row = _next_row_dict()
+                next_row = self._next_row_dict(db_cursor=db_cursor)
                 if next_row is None:
-                    yield {}
+                    # Nothing to yield
+                    yield from ()
                 else:
                     yield next_row
                     # _next_row_dict will return None when at the end
                     # so the sentinel is met by the iter function
-                    yield from iter(_next_row_dict, None)
+                    yield from iter(lambda: self._next_row_dict(db_cursor=db_cursor), None)
 
-        except sqlite3.OperationalError as ex:
+        except sqlite3.Error as ex:
             message = (
-                f"While executing SQL query {sql_query=} to [{self._filename}] operation failed!"
-                f" traceback: {repr(ex)}"
+                f"While executing SQL query {sql_query=} on host [{self._filename}] operation"
+                f" failed! traceback: {repr(ex)}"
             )
             module_logger.error(message)
-            raise exceptions.SQLiteError(message)
+            raise exceptions.SQLiteError(message) from None
 
         finally:
             db_cursor.close()
             self.close_db_connection()
+
+    def _next_row_dict(self, db_cursor: sqlite3.Cursor) -> Optional[dict[str, Any]]:
+        """
+        Returns the next row as a dictionary or None if no
+        more rows.
+
+        :return: The next row as a dictionary or None.
+        """
+
+        row_fetched = db_cursor.fetchone()
+        if row_fetched is None:
+            return None
+        else:
+            return dict(row_fetched)
