@@ -56,7 +56,7 @@ __all__ = [
 module_logger = logging.getLogger(__name__)
 
 # Type aliases
-#
+RedisClient = redis.client.Redis
 
 
 class RedisCacheManager:
@@ -103,10 +103,10 @@ class RedisCacheManager:
 
         self._categories = set(category_keys)
         self._redis_cached_client: Optional[redis.client.Redis] = None
-        self._serializer = RedisSerializer()
+        self._serializer = _RedisSerializer()
 
     @property
-    def _redis_client(self) -> redis.client.Redis:
+    def _redis_client(self) -> RedisClient:
         """
         Lazily initializes and returns the Redis client.
 
@@ -124,7 +124,26 @@ class RedisCacheManager:
 
         return self._redis_cached_client
 
-    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, tries=2, delay_secs=1)
+    def invalidate_client_cache(self) -> None:
+        """
+        Clears the cached client.
+
+        This method allows manually invalidating the cached client,
+        forcing a new client instance to be created on the next access.
+
+        :return: None.
+        :raise RedisCacheManagerError: Raises an error if caching is not
+            enabled for this instance.
+        """
+
+        if not self._redis_cached_client:
+            raise exceptions.RedisCacheManagerError(
+                f"Session caching is not enabled for this instance of {self.__class__.__qualname__}"
+            )
+
+        self._redis_cached_client = None
+
+    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, delay_secs=1)
     def has(self, category: str, key: str) -> bool:
         """
         Checks if a key exists in the cache.
@@ -136,24 +155,23 @@ class RedisCacheManager:
             accessing Redis.
         """
 
-        if category in self._categories:
+        if category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
+        try:
             redis_key = self._serializer.serialize_redis_key(category, key)
 
-            try:
-                redis_response = self._redis_client.exists(redis_key)
-
-            except Exception as ex:
-                raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
+            redis_response = self._redis_client.exists(redis_key)
 
             # Returns the number of keys that exist
             exists = redis_response == 1
 
             return exists
 
-        else:
-            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+        except Exception as ex:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
 
-    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, tries=2, delay_secs=1)
+    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, delay_secs=1)
     def get(self, category: str, key: str) -> Optional[Any]:
         """
         Retrieves a value from the cache.
@@ -165,28 +183,26 @@ class RedisCacheManager:
             accessing Redis.
         """
 
-        if category in self._categories:
+        if category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
+        try:
             redis_key = self._serializer.serialize_redis_key(category, key)
 
-            try:
-                redis_response = self._redis_client.get(redis_key)
-
-            except Exception as ex:
-                raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
+            redis_response = self._redis_client.get(redis_key)
 
             if redis_response:
                 assert isinstance(redis_response, str)
                 response = self._serializer.deserialize(redis_response)
-
             else:
                 response = None
 
             return response
 
-        else:
-            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+        except Exception as ex:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
 
-    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, tries=2, delay_secs=1)
+    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, delay_secs=1)
     def set(self, category: str, key: str, value: Any) -> bool:
         """
         Sets a value in the cache.
@@ -200,27 +216,23 @@ class RedisCacheManager:
             accessing Redis.
         """
 
-        if category in self._categories:
+        if category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
+        try:
             redis_key = self._serializer.serialize_redis_key(category, key)
             redis_value = self._serializer.serialize(value)
 
-            try:
-                redis_response = self._redis_client.set(redis_key, redis_value)
+            redis_response = self._redis_client.set(redis_key, redis_value)
 
-                if redis_response:
-                    assert isinstance(redis_response, bool)
-                    return redis_response
+            response = bool(redis_response)
 
-                else:
-                    return False
+            return response
 
-            except Exception as ex:
-                raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
+        except Exception as ex:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
 
-        else:
-            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
-
-    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, tries=2, delay_secs=1)
+    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, delay_secs=1)
     def delete(self, category: str, key: str) -> bool:
         """
         Invalidates a specific key in the cache.
@@ -233,22 +245,21 @@ class RedisCacheManager:
             accessing Redis.
         """
 
-        if category in self._categories:
+        if category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
+        try:
             redis_key = self._serializer.serialize_redis_key(category, key)
 
-            try:
-                redis_response = self._redis_client.delete(redis_key)
-
-            except Exception as ex:
-                raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
+            redis_response = self._redis_client.delete(redis_key)
 
             # Returns the number of keys invalidated
             invalidated = redis_response == 1
 
             return invalidated
 
-        else:
-            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+        except Exception as ex:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
 
     def clear(self, category: Optional[str] = None) -> bool:
         """
@@ -262,15 +273,11 @@ class RedisCacheManager:
             accessing Redis.
         """
 
+        if category and category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
         if category:
-            if category in self._categories:
-                categories = {category}
-
-            else:
-                raise exceptions.RedisCacheManagerError(
-                    f"[CACHE] - Category {category} not recognized"
-                )
-
+            categories = {category}
         else:
             categories = self._categories
 
@@ -278,21 +285,14 @@ class RedisCacheManager:
         running_total = 0
 
         for cat in categories:
-            # Check there are keys in the cache
-            redis_keys = self.get_keys(category=cat)
-            if redis_keys is None:
-                continue
-
-            for redis_key in redis_keys:
+            for redis_key in self.get_keys(category=cat):
                 keys_to_delete += 1
                 if self.delete(category=cat, key=redis_key):
                     running_total += 1
 
-        if running_total == keys_to_delete:
-            return True
+        response = running_total == keys_to_delete
 
-        else:
-            return False
+        return response
 
     def keys_count(self, category: Optional[str] = None) -> int:
         """
@@ -308,95 +308,47 @@ class RedisCacheManager:
             accessing Redis.
         """
 
+        if category and category not in self._categories:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
+
         if category:
-            if category in self._categories:
-                categories = {category}
-
-            else:
-                raise exceptions.RedisCacheManagerError(
-                    f"[CACHE] - Category {category} not recognized"
-                )
-
+            categories = {category}
         else:
             categories = self._categories
 
         running_total = 0
 
         for cat in categories:
-            # Check there are keys in the cache
-            redis_keys = self.get_keys(category=cat)
-            if redis_keys is None:
-                continue
-
-            for _ in redis_keys:
+            for _ in self.get_keys(category=cat):
                 running_total += 1
 
         return running_total
 
-    def get_category(self, category: str) -> Optional[Iterator[tuple[str, Any]]]:
-        """
-        Retrieves key-value pairs for all items in the specified cache
-        category.
-
-        :param category: The cache category.
-        :return: An iterator of the cached key-value pairs, or None if
-            no items in cache.
-        :raise: RedisCacheManagerError if an error occurs while
-            accessing Redis.
-        """
-
-        redis_keys = self.get_keys(category=category)
-        redis_values = self.get_values(category=category)
-
-        if redis_keys is None or redis_values is None:
-            return None
-
-        response = zip(redis_keys, redis_values)
-
-        return response
-
-    @utils.retry(exception_to_check=exceptions.RedisCacheManagerError, tries=2, delay_secs=1)
-    def get_keys(self, category: str) -> Optional[Iterator[str]]:
+    def get_keys(self, category: str) -> Generator[str, None, None]:
         """
         Retrieves all keys in the specified cache category.
 
         :param category: The cache category.
-        :return: An iterator of the cached keys, or None if no keys
-            in cache.
+        :return: An iterator of the cached keys.
         :raise: RedisCacheManagerError if an error occurs while
             accessing Redis.
         """
 
-        if category in self._categories:
-            redis_key_pattern = self._serializer.serialize_redis_key(category)
-
-            try:
-                redis_response = self._redis_client.scan_iter(match=redis_key_pattern)
-
-            except Exception as ex:
-                raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
-
-            try:
-                first_value = next(redis_response)
-
-            except StopIteration:
-                return None
-
-            def gen_wrapper() -> Generator[str, None, None]:
-                yield first_value.split(':', 1)[1]
-
-                for value in redis_response:
-                    yield value.split(':', 1)[1]
-
-            # This is the iterator we are going to return
-            keys = gen_wrapper()
-
-            return keys
-
-        else:
+        if category not in self._categories:
             raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
 
-    def get_values(self, category: str) -> Optional[Iterator[Any]]:
+        try:
+            redis_key_pattern = self._serializer.serialize_redis_key(category)
+
+            with utils.retry(exception_to_check=Exception, delay_secs=1) as retryer:
+                redis_response = retryer(self._redis_client.scan_iter, match=redis_key_pattern)
+
+            yield from (value.split(':', 1)[1] for value in redis_response)
+
+        except Exception as ex:
+            raise exceptions.RedisCacheManagerError(f"[CACHE] - Redis error: {str(ex)}")
+
+    def get_values(self, category: str) -> Iterator[Any]:
         """
         Retrieves all values in the specified cache category.
 
@@ -407,21 +359,175 @@ class RedisCacheManager:
             accessing Redis.
         """
 
-        if category in self._categories:
-            redis_keys = self.get_keys(category=category)
-
-            if redis_keys is None:
-                return None
-
-            response = (self.get(category=category, key=redis_key) for redis_key in redis_keys)
-
-            return response
-
-        else:
+        if category not in self._categories:
             raise exceptions.RedisCacheManagerError(f"[CACHE] - Category {category} not recognized")
 
+        response = (
+            self.get(category=category, key=redis_key)
+            for redis_key in self.get_keys(category=category)
+        )
 
-class RedisSerializer:
+        return response
+
+    def get_category(self, category: str) -> Iterator[tuple[str, Any]]:
+        """
+        Retrieves key-value pairs for all items in the specified cache
+        category.
+
+        :param category: The cache category.
+        :return: An iterator of the cached key-value pairs.
+        :raise: RedisCacheManagerError if an error occurs while
+            accessing Redis.
+        """
+
+        redis_keys = self.get_keys(category=category)
+        redis_values = self.get_values(category=category)
+
+        response = zip(redis_keys, redis_values)
+
+        return response
+
+
+class _RedisEncoder(json.JSONEncoder):
+    """
+    _RedisEncoder extends the JSONEncoder to support additional
+    data types.
+    """
+
+    def default(self, obj: Any) -> Any:
+        """
+        Override the default method to handle additional data types.
+
+        A function that gets called for objects that can’t otherwise
+        be serialized. It should return a JSON encodable version of
+        the object.
+
+        :param obj: The object to be serialized.
+        :return: The serialized form of the object.
+        """
+
+        if isinstance(obj, (set, tuple, bytes, complex)):
+            return self._default(obj)
+
+        else:
+            raise exceptions.RedisCacheManagerError(
+                f"Object of type {type(obj)!r} for {obj!r} is not supported by Redis serialization."
+            )
+
+    def _default(self, obj: Any) -> Any:
+        """
+        Custom method to handle serialization recursively of
+        specific types.
+
+        :param obj: The object to be serialized.
+        :return: The serialized form of the object.
+        """
+
+        if isinstance(obj, set):
+            return {'__sentinel_type__': 'set', 'data': [self._default(el) for el in obj]}
+
+        elif isinstance(obj, bytes):
+            return {'__sentinel_type__': 'bytes', 'data': obj.decode('utf-8')}
+
+        elif isinstance(obj, complex):
+            return {'__sentinel_type__': 'complex_num', 'data': [obj.real, obj.imag]}
+
+        # Note to future carlogtt, tuple does not call the default
+        # function as they are not seen as unserializable obj, but
+        # they are seen as list
+        elif isinstance(obj, tuple):
+            return {'__sentinel_type__': 'tuple', 'data': [self._default(el) for el in obj]}
+
+        else:
+            return obj
+
+    def encode(self, obj: Any) -> str:
+        """
+        Return a JSON string representation of a Python data
+        structure.
+
+        :param obj: The object to be serialized.
+        :return: The JSON string representation of the object.
+        """
+
+        # Handle tuples because JSON is converting them to lists
+        # Note to future carlogtt, not sure set will ever get here
+        if isinstance(obj, (dict, set, list, tuple)):
+            obj = self._encode(obj)
+
+        return super().encode(o=obj)
+
+    def _encode(self, obj: Any) -> Any:
+        """
+        Custom method to recursively encode tuple data type.
+
+        :param obj: The object to be serialized.
+        :return: The encoded form of the object.
+        """
+
+        # The super encode method would convert tuples to lists, so
+        # we need to intercept that and convert tuples to dict with
+        # a sentinel_value
+        # Also dicts and lists can contain tuples
+        if isinstance(obj, dict):
+            obj = {key: self._encode(value) for key, value in obj.items()}
+
+        elif isinstance(obj, list):
+            obj = [self._encode(el) for el in obj]
+
+        elif isinstance(obj, tuple):
+            obj = {'__sentinel_type__': 'tuple', 'data': [self._encode(el) for el in obj]}
+
+        # Note to future carlogtt, not sure set will ever get here
+        # as they are converted to dict in the above default method
+        elif isinstance(obj, set):
+            obj = {self._encode(el) for el in obj}
+
+        return obj
+
+
+class _RedisDecoder(json.JSONDecoder):
+    """
+    _RedisDecoder extends the JSONDecoder to support the
+    deserialization of additional data types encoded by the
+    _RedisEncoder.
+    """
+
+    def __init__(self, *args, **kwargs) -> None:
+        super().__init__(object_hook=self.object_hook, *args, **kwargs)
+
+    @staticmethod
+    def object_hook(dict_obj: dict[str, Any]) -> Any:
+        """
+        object_hook is an optional function that will be called with
+        the result of any object literal decoded (a dict).
+        The return value of object_hook will be used instead of the
+        dict.
+
+        Custom object hook to convert dictionaries back to original
+        data types based on sentinel type.
+
+        :param dict_obj: The dictionary to be converted.
+        :return: The deserialized object.
+        """
+
+        if '__sentinel_type__' in dict_obj:
+            if dict_obj['__sentinel_type__'] == 'set':
+                return set(dict_obj['data'])
+
+            elif dict_obj['__sentinel_type__'] == 'tuple':
+                return tuple(dict_obj['data'])
+
+            elif dict_obj['__sentinel_type__'] == 'bytes':
+                return dict_obj['data'].encode('utf-8')
+
+            elif dict_obj['__sentinel_type__'] == 'complex':
+                return complex(dict_obj['data'][0], dict_obj['data'][1])
+
+        return dict_obj
+
+
+class _RedisSerializer:
     """
     RedisSerializer is a utility class for serializing and deserializing
     complex Python data types to JSON format and back.
@@ -434,144 +540,6 @@ class RedisSerializer:
     storage and then back to Python objects when retrieved.
     """
 
-    class _RedisEncoder(json.JSONEncoder):
-        """
-        _RedisEncoder extends the JSONEncoder to support additional
-        data types.
-        """
-
-        def default(self, obj: Any) -> Any:
-            """
-            Override the default method to handle additional data types.
-
-            A function that gets called for objects that can’t otherwise
-            be serialized. It should return a JSON encodable version of
-            the object.
-
-            :param obj: The object to be serialized.
-            :return: The serialized form of the object.
-            """
-
-            if isinstance(obj, (set, tuple, bytes, complex)):
-                return self._default(obj)
-
-            else:
-                raise exceptions.RedisCacheManagerError(
-                    f"Object of type {type(obj)!r} for {obj!r} is not supported by Redis"
-                    " serialization."
-                )
-
-        def _default(self, obj: Any) -> Any:
-            """
-            Custom method to handle serialization recursively of
-            specific types.
-
-            :param obj: The object to be serialized.
-            :return: The serialized form of the object.
-            """
-
-            if isinstance(obj, set):
-                return {'__sentinel_type__': 'set', 'data': [self._default(el) for el in obj]}
-
-            elif isinstance(obj, bytes):
-                return {'__sentinel_type__': 'bytes', 'data': obj.decode('utf-8')}
-
-            elif isinstance(obj, complex):
-                return {'__sentinel_type__': 'complex_num', 'data': [obj.real, obj.imag]}
-
-            # Note to future carlogtt, tuple does not call the default
-            # function as they are not seen as unserializable obj, but
-            # they are seen as list
-            elif isinstance(obj, tuple):
-                return {'__sentinel_type__': 'tuple', 'data': [self._default(el) for el in obj]}
-
-            else:
-                return obj
-
-        def encode(self, obj: Any) -> str:
-            """
-            Return a JSON string representation of a Python data
-            structure.
-
-            :param obj: The object to be serialized.
-            :return: The JSON string representation of the object.
-            """
-
-            # Handle tuples because JSON is converting them to lists
-            # Note to future carlogtt, not sure set will ever get here
-            if isinstance(obj, (dict, set, list, tuple)):
-                obj = self._encode(obj)
-
-            return super().encode(o=obj)
-
-        def _encode(self, obj: Any) -> Any:
-            """
-            Custom method to recursively encode tuple data type.
-
-            :param obj: The object to be serialized.
-            :return: The encoded form of the object.
-            """
-
-            # The super encode method would convert tuples to lists, so
-            # we need to intercept that and convert tuples to dict with
-            # a sentinel_value
-            # Also dicts and lists can contain tuples
-            if isinstance(obj, dict):
-                obj = {key: self._encode(value) for key, value in obj.items()}
-
-            elif isinstance(obj, list):
-                obj = [self._encode(el) for el in obj]
-
-            elif isinstance(obj, tuple):
-                obj = {'__sentinel_type__': 'tuple', 'data': [self._encode(el) for el in obj]}
-
-            # Note to future carlogtt, not sure set will ever get here
-            # as they are converted to dict in the above default method
-            elif isinstance(obj, set):
-                obj = {self._encode(el) for el in obj}
-
-            return obj
-
-    class _RedisDecoder(json.JSONDecoder):
-        """
-        _RedisDecoder extends the JSONDecoder to support the
-        deserialization of additional data types encoded by the
-        _RedisEncoder.
-        """
-
-        def __init__(self, *args, **kwargs) -> None:
-            super().__init__(object_hook=self.object_hook, *args, **kwargs)
-
-        @staticmethod
-        def object_hook(dict_obj: dict[str, Any]) -> Any:
-            """
-            object_hook is an optional function that will be called with
-            the result of any object literal decoded (a dict).
-            The return value of object_hook will be used instead of the
-            dict.
-
-            Custom object hook to convert dictionaries back to original
-            data types based on sentinel type.
-
-            :param dict_obj: The dictionary to be converted.
-            :return: The deserialized object.
-            """
-
-            if '__sentinel_type__' in dict_obj:
-                if dict_obj['__sentinel_type__'] == 'set':
-                    return set(dict_obj['data'])
-
-                elif dict_obj['__sentinel_type__'] == 'tuple':
-                    return tuple(dict_obj['data'])
-
-                elif dict_obj['__sentinel_type__'] == 'bytes':
-                    return dict_obj['data'].encode('utf-8')
-
-                elif dict_obj['__sentinel_type__'] == 'complex':
-                    return complex(dict_obj['data'][0], dict_obj['data'][1])
-
-            return dict_obj
-
     def serialize(self, obj: Any) -> str:
         """
         Serialize a Python object to a JSON string using the custom
@@ -581,7 +549,7 @@ class RedisSerializer:
         :return: The JSON string representation of the object.
         """
 
-        return json.dumps(obj=obj, cls=self._RedisEncoder)
+        return json.dumps(obj=obj, cls=_RedisEncoder)
 
     def deserialize(self, s: str) -> Any:
         """
@@ -592,7 +560,7 @@ class RedisSerializer:
         :return: The deserialized Python object.
         """
 
-        return json.loads(s=s, cls=self._RedisDecoder)
+        return json.loads(s=s, cls=_RedisDecoder)
 
     @staticmethod
     def serialize_redis_key(category: str, key: Optional[str] = None) -> str:
